import subprocess

from .color import ColorClass

Color = ColorClass()

#------------------------------
# Confirm the existence of the Git repository.
#------------------------------

class CheckClass:
    #def __init__(self):

    def do_command(self,command):
        proc = subprocess.Popen(
            command,
            shell  = True,
            stdin  = subprocess.PIPE,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE)
        stdout_data, stderr_data = proc.communicate()
        return stdout_data.decode()

    def repository_exists(self, path = '.'):
        command = "git -C %s status -s 2>&1 > /dev/null | awk '{print $1}'" % path
        if self.do_command(command) == '':
            return True
        else:
            return False

    def stage_exists(self, path = '.'):
        command = "git -C %s status -s" % path
        if self.do_command(command) == '':
            return False
        else:
            return True

    def get_git_status(self, path = '.'):
        command = "git -C %s status -s" % path
        status  = self.do_command(command)
        if not status:
            return ''
        else:
            return status

    def is_change(self, area=''):
        workingtree = ''
        index       = ''
        if self.get_git_status() == False:
            return False
        status = self.get_git_status().split('\n')
        for item in status:
            if item:
                index       = index + item.rsplit(' ')[0]
                workingtree = workingtree + item.rsplit(' ')[1]

        if area == 'index':
            if index:
                return True
            else:
                return False
        elif area == 'workingtree':
            if workingtree:
                return True
            else:
                return False
        else:
            return False

    def get_git_branch(self, path = '.'):
        command = "git -C %s rev-parse --abbrev-ref HEAD" % path
        branch  = self.do_command(command).replace('\n', '')
        return branch

    def branch_exists(self, path = '.'):
        command = "git -C %s branch" % path
        branch  = self.do_command(command).replace('\n', '')
        return branch

    def commit_exists(self, path = '.'):
        branch = self.get_git_branch().rstrip('\r\n')
        command = 'git -C %s log --pretty=format:"%%H" origin/%s..HEAD' % (path, branch)
        commit  = self.do_command(command)
        if not commit:
            return False
        else:
            return commit

    def get_unpushed_list(self, path = '.'):
        branch = self.get_git_branch().rstrip('\r\n')
        command1 = 'git -C %s log --pretty=format:"%%h" origin/%s..HEAD' % (path, branch)
        command2 = 'git -C %s log --pretty=format:"%%d" origin/%s..HEAD' % (path, branch)
        short_hash = self.do_command(command1)
        ref_name   = self.do_command(command2)
        if not short_hash:
            return False
        else:
            return short_hash + ref_name

    def get_latest_tag(self, path = '.'):
        command = 'git -C %s tag | sed s/v//g | sort -t . -n -k1,1 -k2,2 -k3,3 | tail -n1' % path
        tag     = self.do_command(command).replace('\n','')
        return tag

    def do_git_tag(self, tag = 0, path = '.'):
        command = 'git -C %s tag -a v%s -m "%s"' % (path, tag, tag)
        tag     = self.do_command(command)
        return tag

    def do_git_add(self, path = '.'):
        command = 'git -C %s add -A' % path
        result  = self.do_command(command)
        return result

    def do_git_commit(self, message = '', path = '.'):
        message = message.replace('\n','')
        command = 'git -C %s commit -m "%s"' % (path, message)
        result  = self.do_command(command)
        return result

    def do_git_push(self, branch = 'main', path = '.'):
        command = 'git -C %s push origin %s' % (path, branch)
        push    = self.do_command(command)
        return push

    """
    see also: https://git-scm.com/docs/git-status
    """
    def status(self, status = ''):
        index_i       = 0
        workingtree_i = 0
        status = self.get_git_status()
        status_dict = {
            'M  ':'i.Modified ',
            'MM ':'i.Modified*',
            'MD ':'i.Modified* ',
            'A  ':'i.Added ',
            'AM ':'i.Added* ',
            'AD ':'i.Added* ',
            'D  ':'i.Deleted ',
            'R  ':'i.Renamed ',
            'RM ':'i.Renamed* ',
            'RD ':'i.Renamed* ',
            'C  ':'i.Copied ',
            'CM ':'i.Copied* ',
            'CD ':'i.Copied* ',
            ' M ':'w.Updated ',
            'MM ':'w.Updated* ',
            'AM ':'w.Updated* ',
            'RM ':'w.Updated* ',
            'CM ':'w.Updated* ',
            ' D ':'w.Deleted ',
            'MD ':'w.Deleted* ',
            'AD ':'w.Deleted* ',
            'RD ':'w.Deleted* ',
            'CD ':'w.Deleted* ',
            ' R ':'w.Renamed ',
            'DR ':'w.Renamed* ',
            ' C ':'w.Copied ',
            'DC ':'w.Copied* ',
            'DD ':'o.Unmerged ',
            'AU ':'o.Unmerged ',
            'UD ':'o.Unmerged ',
            'UA ':'o.Unmerged ',
            'DU ':'o.Unmerged ',
            'AA ':'o.Unmerged ',
            'UU ':'o.Unmerged ',
            '?? ':'o.Untracked ',
            '!! ':'o.Ignored ',
        }
        for word, read in status_dict.items():
            status = status.replace(word, read)
        status_list = status.split('\n')
        del status_list[-1]
        # Border between status display.
        if not status_list:
            max_file_name = 1
        else:
            max_file_name = len(max((x for x in status_list), key=len))
        if max_file_name <= 14:
            border = '-' * 17
        else:
            border = '-' * (max_file_name + 2)

        Color.set(self.get_git_branch(), 'green', '', 'bold', '')

        if self.get_latest_tag() == '':
            Color.set(' (no tags)', 'white', '', 'em')
        else:
            Color.set(' (' + self.get_latest_tag() + ')', 'white', '', 'em')

        print(border)

        Color.set(' Working tree ', 'purple', '', 'bold')
        for item in status_list:
            if item.startswith('w.') or item.startswith('o.'):
                if item.startswith('w.'):
                    print(' - ' + item.strip('w.'))
                    workingtree_i += 1
                if item.startswith('o.'):
                    print(' - ', end='')
                    Color.set(item.strip('o.'), 'red')
                    workingtree_i += 1
        if workingtree_i == 0:
            Color.set(' - No files.', 'white', '', 'em')

        Color.set(' Index ', 'purple', '', 'bold')
        it = iter(status_list)
        for item in status_list:
            if item.startswith('i.'):
                print(' - ' + item.strip('i.'))
                index_i += 1
        if index_i == 0:
            Color.set(' - No files.', 'white', '', 'em')

        Color.set(' Unpushed commit ', 'purple', '', 'bold')
        if self.get_unpushed_list():
            print(self.get_unpushed_list())
        else:
            Color.set(' - No commits.', 'white', '', 'em')

        print(border)

    def format_status(self, status = ''):
        format_dict = {
            'M  ':'Updated ',
            'A  ':'Added ',
            'D  ':'Deleted ',
            'R  ':'Renamed ',
            'C  ':'Copied '
        }
        for word, read in format_dict.items():
            status = status.replace(word, read)
        return status.replace('\n', ', ').strip(', ')



import datetime
import time

from .color import ColorClass
from .check import CheckClass
Color = ColorClass()
Check = CheckClass()

class MainClass:
    #def __init(selft):

    #------------------------------
    # Add
    #------------------------------
    def fast_add(self):
        Color.set('ADD ', 'purple', '', 'bold', '')
        try:
            read = input('files to the index? [n/Y or Enter]')
            if read == 'no' or read == 'NO' or read == 'n' or read == 'N':
                exit()
            elif read == 'yes' or read == 'YES' or read == 'y' or read == 'Y':
                Check.do_git_add()
                Color.set('STAGING done. ✔', 'green')
                print()
            else:
                Check.do_git_add()
                Color.set('STAGING done. ✔', 'green')
                print()
        except (KeyboardInterrupt) as e:
            Color.set('Process aborted.', 'red')
            exit()

    #------------------------------
    # Commit
    #------------------------------
    def fast_commit(self, input_message = ''):

        if Check.is_change('workingtree') == True:
            self.fast_add()

        # If there is no commit message argument, the date and status will be used as the message.
        if input_message == '':
            now     = str(datetime.datetime.today())
            message = Check.format_status(Check.get_git_status())
        else:
            message = input_message

        # Push to master if it appears to be the first push where no local branch exists.
        if Check.branch_exists() == '':
            branch = 'main'
            Color.set('Initial commit.', 'yellow')
        else:
            branch = Check.get_git_branch()

        Color.set('COMMIT MESSAGE: ', 'yellow', '', '', '')
        print (message)
        Color.set('BRANCH: ', 'yellow', '', '', '')
        print (branch)

        try:
            Color.set('COMMIT ', 'purple', '', 'bold', '')
            read = input('the index contents? [n/Y or Enter]')
            if read == 'no' or read == 'NO' or read == 'n' or read == 'N':
                exit()
            elif read == 'yes' or read == 'YES' or read == 'y' or read == 'Y':
                Check.do_git_commit(message)
                Color.set('COMMIT done. ✔', 'green')
                print()
            else:
                Check.do_git_commit(message)
                Color.set('COMMIT done. ✔', 'green')
                print()
        except (KeyboardInterrupt) as e:
            Color.set('Process aborted.', 'red')
            exit()

    #------------------------------
    # Push
    #------------------------------
    def fast_push(self, input_message = ''):
        Check.status()
        # Check any file exists in the working tree.
        if Check.is_change('workingtree') == True:
            self.fast_add()
        # Check any file exists in the index. 
        if Check.is_change('index') == True:
            self.fast_commit(input_message)

        # コミット済みファイルが存在するか
        # Check any file exists in the index. 
        if Check.commit_exists() == False:
            exit()

        # ローカルのブランチが存在しない初回pushとみられる場合はmasterにpushする
        if Check.branch_exists == 0:
            branch = 'main'
            Color.set('Initial commit', 'yellow')
        else:
            branch = Check.get_git_branch() 

        try:
            Color.set('BRANCH: ', 'yellow', '', '', '')
            print (branch)
            Color.set('PUSH ', 'purple', '', 'bold', '')
            read = input('local commits? [n/Y or Enter]')
            if read == 'no' or read == 'NO' or read == 'n' or read == 'N':
                exit()
            elif read == 'yes' or read == 'YES' or read == 'y' or read == 'Y':
                push = Check.do_git_push(branch)
                Color.set('PUSH done. ✔', 'green')
            else:
                push = Check.do_git_push(branch)
                Color.set('PUSH done. ✔', 'green')
                exit()
        except (KeyboardInterrupt) as e:
            Color.set('Process aborted.', 'red')
            exit()

    #------------------------------
    # Tags
    #------------------------------
    def fast_tag(self, new_tag = ''):
        latest_tag = Check.get_latest_tag()
        if latest_tag == '':
            latest = '0.0.0'
        else:
            latest = Check.get_latest_tag()
        try:
            latest_list  = latest.split('.')
            latest_major = latest_list[0]
            latest_minor = latest_list[1]
            latest_patch = latest_list[2]
            patch_increment = int(latest_patch) + 1
            patch_ver       = latest_major + '.' + latest_minor + '.' + str(patch_increment)
        except:
            Color.set('Latest tag ', 'red', '', '', '')
            Color.set(latest_tag, 'red', '', 'bold', '')
            Color.set(' is not an auto-incrementable version notation.', 'red')
            Color.set('Supports tags in X.Y.Z format.', 'red')
            patch_ver = '0.0.1'
        if new_tag == '':
            if Check.get_latest_tag() == '':
                Color.set('No tags are currently.', 'green')
            else:
                Color.set('Latest tag is ' + latest, 'green')
            Color.set('Auto incremented version is ' + patch_ver, 'green')
            try:
                Color.set('TAG ', 'purple', '', 'bold', '')
                read = input('the latest commit? [n/Y or Enter]')
                if read == 'no' or read == 'NO' or read == 'n' or read == 'N':
                    exit()
                elif read == 'yes' or read == 'YES' or read == 'y' or read == 'Y':
                    Check.do_git_tag(patch_ver)
                    Check.do_command('git push origin --tag')
                else:
                    Check.do_git_tag(patch_ver)
                    Check.do_command('git push origin --tag')
            except (KeyboardInterrupt) as e:
                Color.set('Process aborted.', 'red')
                exit()
        else: # user add a tag
            if Check.get_latest_tag() == '':
                Color.set('No tags are currently.', 'green')
            else:
                Color.set('Latest tag is ' + latest, 'green')
            Color.set('New tag is ' + new_tag, 'green')
            try:
                read = input('Tagging? [n/Y or Enter]')
                if read == 'no' or read == 'NO' or read == 'n' or read == 'N':
                    exit()
                elif read == 'yes' or read == 'YES' or read == 'y' or read == 'Y':
                    Check.do_git_tag(new_tag)
                    Check.do_command('git push origin --tag')
                else:
                    Check.do_git_tag(new_tag)
                    Check.do_command('git push origin --tag')
            except (KeyboardInterrupt) as e:
                Color.set('Process aborted.', 'red')
                exit()

    #------------------------------
    # How to use
    #------------------------------
    def usage(self):
        Color.set(' HELP ', 'green', 'white')
        Color.set('gnow command :: Git commit for NOW', 'green')
        print("This is the 'gnow' command to do a quick git commit and git push.")
        print('')
        print("Options:")
        print("  (no args): Run git add -> commit -> push")
        print("  -h, --help :Show helps.")
        print("  -v, -V, --version :Show version.")
        print("  -c, --commit [ARG] :Commit only.")
        print("  -t, --tag [ARG] :Add tag. If no argument, current version will be automatically incremented.")




#!/usr/bin/env python3
import sys
import re

from .color import ColorClass
from .check import CheckClass
from .main import MainClass

Color = ColorClass()
Check = CheckClass()
Main  = MainClass()

#------------------------------
# Command arguments
#------------------------------

def run():
    if len(sys.argv) >= 2:
        arg1 = sys.argv[1]
    else:
        arg1 = 0
    if len(sys.argv) >= 3:
        arg2 = sys.argv[2]
    else:
        arg2 = 0
    if len(sys.argv) >= 4:
        arg3 = sys.argv[3]
    else:
        arg3 = 0

    # Help
    if arg1 == '-h' or arg1 == '--help':
        Main.usage()
        exit()
    # Version
    elif arg1 == '-v' or arg1 == '-V' or arg1 == '--version':
        print('0.0.1')
        exit()
    # Commit only
    elif arg1 == '-c' or arg1 == '--commit':
        if Check.repository_exists() == False:
            Color.set('No git repository exists in this directory.', 'red')
            exit()
        if arg2 == 0 or re.compile('^-.+').search(arg2):
            Main.fast_commit()
        else:
            Main.fast_commit(arg2)
    # Tagging
    elif arg1 == '-t' or arg1 == '--tag':
        if Check.repository_exists() == False:
            Color.set('No git repository exists in this directory.', 'red')
            exit()
        if arg2 == 0 or re.compile('^-.+').search(arg2):
            Main.fast_tag()
        else:
            Main.fast_tag(arg2)
    # Status
    elif arg1 == '-s' or arg1 == '--status':
        if Check.repository_exists() == False:
            Color.set('No git repository exists in this directory.', 'red')
            exit()
        Check.status()
    elif arg1 == '-' or arg1 == '--':
        Color.set("Illegal option: Only '-' or '--'",'red')
        exit()
    # Undefined option
    elif arg1 != 0 and re.compile('^-.*').search(arg1):
        Color.set("Illegal option: Undefined option '" + arg1 + "'",'red')
    # No arguments
    elif arg1 != 0 and re.compile('^(?!-.+)').search(arg1):
        if Check.repository_exists() == False:
            Color.set('No git repository exists in this directory.', 'red')
            exit()
        Main.fast_push(arg1)
    # Without any arguments
    elif arg1 == 0:
        if Check.repository_exists() == False:
            Color.set('No git repository exists in this directory.', 'red')
            exit()
        Main.fast_push()
    else:
        exit()

if __name__ == '__main__':
    run()



class ColorClass:
  def __init__(self):
    self.colors = {
      'gray'   : '\033[30',
      'red'    : '\033[31',
      'green'  : '\033[32',
      'yellow' : '\033[33',
      'blue'   : '\033[34',
      'purple' : '\033[35',
      'lime'   : '\033[36',
      'white'  : '\033[37',
    }

    self.backgrounds = {
        'gray'   : ';40',
        'red'    : ';41',
        'green'  : ';42',
        'yellow' : ';43',
        'blue'   : ';44',
        'purple' : ';45',
        'lime'   : ';46',
        'white'  : ';47',
    }

    self.style = {
        'bold'      : ';1',
        'em'        : ';2',
        'underline' : ';4',
        'blink'     : ';5',
        'invert'    : ';7',
        'hidden'    : ';8',
    }

    self.end    = '\033[0m'

  def set(self, text, color, background = '', style = '', end='\n'):
    colorCode = self.colors.get(color, '\033[30m')
    bgCode    = self.backgrounds.get(background, '')
    styleCode = self.style.get(style, '')
    output    = colorCode + bgCode + styleCode + 'm' + text + self.end
    print(output, end=end)




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='gnow',
    version='1.0.0',
    description='This command makes it easy to git add / commit /push and auto-increment tagging.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/addshlab/gnow',
    author='addshlab',
    license='MIT',
    keywords='git',
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        'Topic :: Software Development :: Version Control :: Git',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'console_scripts':[
            'gnow=gnow.command:run',
        ],
    },
)



# Import main class.
from .check import CheckClass
Check = CheckClass()

def create_dir():
    Check.do_command('mkdir git_test')

def delete_dir():
    Check.do_command('rm -rf git_test')

def test_do_command():
    command = 'echo $(( 1 + 1))'
    result  = Check.do_command(command).replace('\n', '')
    assert int(result) == 2

def test_no_repository_exists():
    Check.do_command('mkdir /tmp/gnow_test')
    result = Check.repository_exists('/tmp/gnow_test')
    Check.do_command('rm -rf /tmp/gnow_test')
    assert result == 0

def test_yes_repository_exists():
    Check.do_command('mkdir /tmp/gnow_test')
    Check.do_command('git init /tmp/gnow_test')
    result = Check.repository_exists('/tmp/gnow_test')
    Check.do_command('rm -rf /tmp/gnow_test')
    assert result == 1




#def test_stage_exists():
#    result = Check.stage_exists()
#    assert result == True

#def test_get_git_status():
#def test_is_change(area=''):
#def test_get_git_branch():
#def test_branch_exists():
#def test_commit_exists():
#def test_get_latest_tag():
#def test_do_git_tag(tag = 0):
#def test_do_git_add():
#def test_do_git_commit(message = ''):
#def test_do_git_push(branch = 'main'):
#def test_status(status = ''):
#def test_format_status(status = ''):
